# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006-2008 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 3.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.

"""
Module responsible for starting the Application
"""

__maintainer__ = 'Philippe Normand <philippe@fluendo.com>'

import sys
import os
import datetime
import gobject
import platform

import dbus
import dbus.service

from elisa.core import __version__, version_info
from elisa.core.config import Config, ConfigError
from elisa.core.log import Loggable
from elisa.core.utils import exception_hook, misc
from elisa.core import bus

from elisa.core.components.message import Message

from elisa.core import config_upgrader
from elisa.core.plugin_registry import PluginRegistry
from elisa.core import input_manager, service_manager
from elisa.core import metadata_manager, resource_manager
from elisa.core import interface_controller
from elisa.core.media_directory_helper import MediaDirectoryHelper

from twisted.internet import reactor, defer, task, threads
from twisted.web import client

# import pkg_resources after plugin_registry
import pkg_resources

# database
try:
    from elisa.extern.storm_wrapper import store
    from storm.locals import create_database
except ImportError:
    store = None

try:
    from elisa.core.utils.update_checker import UpdateChecker
except ImportError:
    # could happen if the elisa http client plugin is missing
    UpdateChecker = None


from distutils.version import LooseVersion

CONFIG_DIR = os.path.join(os.path.expanduser('~'), ".elisa-0.5")

# We add to the configuration file name the version number in order to bypass
# the configuration upgrade system. Direct consequences:
# - custom user configuration is lost between releases
# - prevents configuration upgrade bugs
# - avoid configuration upgrade work
#version_string = "_".join([str(n) for n in version_info])
#CONFIG_FILE = "elisa_%s.conf" % version_string

# FIXME: hardcoded configuration file name in order to avoid its rewriting
# starting from Elisa 0.5.6; this is a temporary solution that is valid for
# the moment given that there is no configuration to upgrade for now.
CONFIG_FILE = "elisa_0_5_6.conf"

TRANS_FILE = "data/translations.lst"
PICTURES_CACHE = os.path.join(CONFIG_DIR, 'pictures_cache')

DEFAULT_CONFIG = """\
[general]
version = '%(version)s'
install_date = '%(install_date)s'
frontends = ['frontend1']
disabled_plugins = []
# database connection string. see https://storm.canonical.com/Manual
database = '%(database_uri)s'

[frontend1]
frontend = 'pigment.pigment_frontend:PigmentFrontend'
theme = 'elisa.plugins.hp'
controller_path = '/poblesec'

[directories]
video = ['*default*']
pictures = ['*default*']
music = ['*default*']

[poblesec.main:PoblesecController]
viewport_buttons_mode = 'embedded'

[poblesec.player_audio:AudioPlayerController]
visualization = ''
osd_timeout = '0'
audio_sink = 'gconfaudiosink'

[poblesec.player_video:VideoPlayerController]
audio_sink = 'gconfaudiosink'
"""

def setup_output_redirection(directory_path, log_filename=None):
    # See http://www.py2exe.org/index.cgi/StderrLog
    log_filename = log_filename or "elisa.log"
    elisa_log = os.path.join(directory_path, log_filename)
    sys.stdout = open(elisa_log, "w")
    sys.stderr = open(elisa_log, "w")


class ComponentsLoadedMessage(Message):
    """
    Sent when all components have been instantiated
    """


class NewElisaVersionMessage(Message):
    """
    Sent when Application detects a new Elisa version is available
    online.

    @ivar version: new version in dotted syntax (ex: 10.0.0)
    @type version: unicode
    @ivar installer_url:     URL of the new win32 installer (optional)
    @type installer_url:     unicode
    """

    def __init__(self, version, installer_url=None):
        super(NewElisaVersionMessage, self).__init__()
        self.version = version
        self.installer_url = installer_url


class FirstRunMessage(Message):
    """
    Sent when Elisa detects it is the first time it is being run.
    """


class BaseApplication(Loggable):
    config = None
    plugin_registry = None
    bus = None
    metadata_manager = None
    service_manager = None
    input_manager = None
    store = None

    def __init__(self, config, plugin_directories):
        # we do this here and not in initialize() as this needs to be done as
        # soon as possible
        super(BaseApplication, self).__init__()

        self._first_run = False
        self._config_filename = self._get_config_filename(config)
        if not os.path.exists(self._config_filename):
            self._first_run = True

        # if running frozen on win32, redirect stdout/stderr to $CONFIG_DIR/elisa.log
        if hasattr(sys, "frozen"):
            setup_output_redirection(os.path.dirname(self._config_filename))

        self.config, self._install_date, self._user_id = self._load_config(config)
        self.plugin_registry = self._create_plugin_registry(plugin_directories)

    def _create_plugin_registry(self, plugin_directories):
        disabled_plugins = self.config.get_option('disabled_plugins',
                                                  section='general',
                                                  default=[])
        plugin_registry = PluginRegistry(plugin_directories)
        plugin_registry.load_plugins(disabled_plugins)
        return plugin_registry

    def _get_config_filename(self, config):
        if config is not None and not isinstance(config, basestring):
            # config is already a config-object
            config_filename = config.get_filename()
        elif config is None:
            config_filename = os.path.join(CONFIG_DIR, CONFIG_FILE)
        else:
            # config is a string
            config_filename = config

        dirname = os.path.dirname(config_filename)
        if dirname and not os.path.exists(dirname):
            try:
                os.makedirs(dirname)
            except OSError, e:
                self.warning("Could not create '%s': %s" % (dirname, e))
                raise
        return config_filename

    def _load_config(self, config):
        today = datetime.date.today().isoformat()
        if config is not None and not isinstance(config, basestring):
            # config is already a config-object
            cfg = config
        else:
            config_filename = self._config_filename
            self.info("Using config file: %r", config_filename)
            default_config = DEFAULT_CONFIG % {'version': __version__,
                    'install_date': today,
                    'database_uri': 'sqlite:%s/elisa.db' % \
                    CONFIG_DIR.replace('\\','/')}

            try:
                cfg = Config(config_filename, default_config=default_config)
            except ConfigError, error:
                self.warning(error)
                raise

            if not cfg.first_load:
                # ok we might have an old config format here
                upgrader = config_upgrader.ConfigUpgrader(cfg, default_config)
                cfg = upgrader.update_for(version_info)

        install_date = cfg.get_option('install_date', section='general',
                                      default=today)
        user_id = cfg.get_option('user_id', section='general', default='')
        return (cfg, install_date, user_id)


class Application(BaseApplication):
    """ Application is the entry point of Elisa. It groups all the necessary
    elements needed for Elisa to run. It is in charge of instantiating a
    Config and a PluginRegistry. Application also provides access to
    input events and data, and holds the user interfaces.
 
    @ivar plugin_registry:      loads and manages the plugins
    @type plugin_registry:      L{elisa.core.plugin_registry.PluginRegistry}
    @ivar config:               Application's configuration file, storing options
    @type config:               L{elisa.core.config.Config}
    @ivar options:              Options passed on the command line when
                                launching elisa
    @type options:              L{elisa.core.options.Options}
    @ivar bus:                  DOCME
    @type bus:                  L{elisa.core.bus.Bus}
    @ivar metadata_manager:     DOCME
    @type metadata_manager:     L{elisa.core.metadata_manager.MetadataManager}
    @ivar resource_manager:     DOCME
    @type resource_manager:     L{elisa.core.resource_manager.ResourceManager}
    @ivar service_manager:      DOCME
    @type service_manager:      L{elisa.core.service_manager.ServiceManager}
    @ivar interface_controller: DOCME
    @type interface_controller: L{elisa.core.interface_controller.InterfaceController}
    @ivar input_manager:        DOCME
    @type input_manager:        L{elisa.core.input_manager.InputManager}
    @ivar store:                the access point to the database using storm
    @type store:                L{elisa.extern.twisted_storm.store.DeferredStore}
    """

    log_category = "application"

    def __init__(self, config_filename=None, show_tracebacks=False,
                 splash=None, options=None):
        """
        @param config_filename: the config filename to use. Can be absolute
                                or relative path
        @type config_filename:  string or None to use default config file
        """
        default_path = os.path.join(CONFIG_DIR, 'plugins')
        paths = os.environ.get('ELISA_PLUGIN_PATH', default_path).split(os.path.pathsep)
        paths = [p for p in paths if p != '']
        if default_path not in paths:
            paths.insert(0, default_path)

        super(Application, self).__init__(config_filename, paths)

        # load the exception hook asap
        self.show_tracebacks = show_tracebacks
        self._load_exception_hook()

        self.debug("Creating")
        self.running = False
        self._splash = splash

        self.store = None
        self.update_checker = None

        self.bus = bus.Bus()
        self.service_manager = service_manager.ServiceManager()
        self.metadata_manager = metadata_manager.MetadataManager()
        self.resource_manager = resource_manager.ResourceManager()
        self.input_manager = input_manager.InputManager()
        self.interface_controller = interface_controller.InterfaceController()
        self.options = options

        if self._first_run:
            # First time elisa is run with this configuration, we want to
            # notify other components (so that the frontend can run a
            # configuration wizard for instance).
            self.bus.send_message(FirstRunMessage())

        import platform
        system = platform.system()
        if system == 'Windows':
            from elisa.plugins.poblesec.windows_message import WindowsMessageSource
            from win32con import WM_COPYDATA
            self.windows_message_source = WindowsMessageSource()
            winprocs = { WM_COPYDATA : self._on_copydata, }
            self.windows_message_source.add_wndproc(winprocs)

    def _update_check_callback(self, results):
        try:
            version = results['version']
        except KeyError:
            version = None

        try:
            user_id = results['user_id']
        except KeyError:
            user_id = None

        try:
            installer_url = results['installer_url']
        except KeyError:
            installer_url = None

        if version and LooseVersion(version) > LooseVersion(__version__):
            msg = NewElisaVersionMessage(version, installer_url)
            self.bus.send_message(msg)

        # store user_id in config
        self.config.set_option('user_id', user_id, section='general')
        self._user_id = user_id

        # store the country code
        try:
            self.config.set_option('country_code', results['country_code'],
                    section='general')
        except KeyError:
            # country code not given in the response
            pass

    def _load_exception_hook(self):
        """ Override the default system exception hook with our own
        """
        # FIXME: make this configurable
        self.logdir = None
        self.debug_level = int(self.show_tracebacks)

        sys.excepthook = self._excepthook
        # log twisted errors: Deactivated
        # from twisted.python import log
        # log.err = self.log_failure

    def _excepthook(self, *args):
        data = exception_hook.format_traceback(*args)
        path = exception_hook.write_to_logfile(data, self.logdir)
        self.warning("An Traceback occurred and got saved to %s" % path)
        self._after_hook(data)

    def _after_hook(self, data):
        if self.debug_level > 0:
            print data
            if self.debug_level == 2:
                try:
                    import pdb
                except ImportError:
                    print "pdb missing. debug shell not started"
                    return

                print "You are now in a debug shell. Application hold until" \
                      " you press 'c'!"
                pdb.set_trace()

    # Exception handling methods
    def log_traceback(self):
        """
        Log the traceback without stopping the process. This could ususally be
        used in parts, where you want to go on and log the exception.
        Example::

            try:
                component.initialize()
            except:
                # and log all the other exceptions
                path = application.log_traceback()
                self.warning("Initilize Component '%s' failed. Traceback saved at %s" % path)
            self.going_on()

        @return: path to the file, where the traceback got logged
        """
        data = exception_hook.format_traceback()
        path = exception_hook.write_to_logfile(data, self.logdir)
        self._after_hook(data)
        return path

    def log_failure(self, failure):
        """
        Log the twisted failure without re-raising the exception. Example in
        an errback::

            def errback(failure):
                path = application.log_failure(failure)
                self.warning("Connection refused. Full output at %s" % path)
                return

        @param failure: the failure to log
        @type failure:  L{twisted.python.failure.Failure}

        @return: path to the file, where the traceback got logged
        """
        data = exception_hook.format_failure(failure)
        path = exception_hook.write_to_logfile(data, self.logdir)
        self._after_hook(data)
        return path

    def initialize(self):
        """
        Load the providers for the different managers, then initialize the
        interface_controller.
        """
        # If the configuration directory does not exist, create it
        if not os.path.exists(CONFIG_DIR):
            os.makedirs(CONFIG_DIR, 0755)

        # we want to load the providers and then initialize the
        # interface controller
        self.media_directories = MediaDirectoryHelper()

        def loading_providers_done(result):
            return self.interface_controller.initialize()

        managers = ( # manager, entry_point
                    (self.resource_manager, 'resource_provider'),
                    (self.service_manager, 'service_provider'),
                    (self.metadata_manager, 'metadata_provider'),
                    (self.input_manager, 'input_provider'),
                   )

        def iter_managers(managers):
            for manager, entry_point in managers:
                entry_point = 'elisa.core.components.' + entry_point
                self.debug("Loading providers for manager %s" % manager)

                providers = []
                for entry in pkg_resources.iter_entry_points(entry_point):
                    # FIXME: consider only "enabled plugins"
                    component_name = '%s:%s' % (entry.module_name,
                            '.'.join(entry.attrs))

                    providers.append(component_name)

                if not providers:
                    self.debug("No providers to load for manager %s" % manager)
                    continue

                dfr = manager.load_components(providers)
                yield dfr
        
        def load_managers(db, managers):
            return task.coiterate(iter_managers(managers)).addCallback(
                    loading_providers_done)

        def interface_initialized(result):
            # Now that db/plugins are loaded, we can safely register dbus service name
            self.dbus_name = dbus.service.BusName('com.fluendo.Elisa', dbus.SessionBus())
            self.bus.send_message(ComponentsLoadedMessage())
            self._close_splash_screen()

        dfr = self.initialize_db()
        dfr.addCallback(load_managers, managers)
        dfr.addCallback(interface_initialized)
        return dfr

    def initialize_db(self):
        """
        initialize the database depending on the configuration
        """
        if store == None:
            self.warning("Could not import storm. Database disabled")
            return defer.succeed(None)

        db_string = self.config.get_option('database', section='general',
                                                    default='')

        try:
            db = create_database(db_string)
        except Exception, e:
            return defer.fail(e)

        self.store = store.DeferredStore(db, auto_reload=False)
        return self.store.start()


    def start(self):
        """ Execute the application. Start the Managers and the
        InterfaceController.
        """
        if hasattr(gobject, 'set_prgname') == True:
            gobject.set_prgname('elisa')

        self.running = True
        self.info("Starting")
        self.bus.start()
        self.input_manager.start()
        self.metadata_manager.start()
        self.resource_manager.start()
        self.service_manager.start()

        if UpdateChecker:
            self.update_checker = UpdateChecker(self._install_date,
                    self._user_id,  __version__)
            self.update_checker.start(self._update_check_callback)

    def stop(self, stop_reactor=True):
        """
        Stop the application.

        @param stop_reactor: stop the reactor after stopping the application
        @type stop_reactor:  C{bool}
        @rtype:              L{twisted.internet.defer.Deferred}
        """
        self._close_splash_screen()

        if self.update_checker is not None:
            self.update_checker.stop()

        def interface_controller_stopped(result):
            self.info("Stopping managers")

            manager_deferreds = []
            for manager in (self.service_manager, self.metadata_manager,
                            self.input_manager, self.resource_manager):
                manager_deferreds.append(defer.maybeDeferred(manager.stop))
            if self.store:
                # FIXME: store is not closed properly: self.store.stop should be
                # called after the commit
                manager_deferreds.append(self.store.commit())

            dfr = defer.DeferredList(manager_deferreds)
            dfr.addCallback(managers_stopped)
            return dfr

        def managers_stopped(managers):
            self.info("Stopping reactor")
            self.bus.stop()

            self.running = False

            if self.config:
                self.config.write()

            if stop_reactor and reactor.running:
                if platform.system() == 'Windows':
                    # FIXME: extremely ugly hack to terminate Elisa under Windows
                    # fix needed:
                    # - fix the locks in gst_metadata when it is out of process
                    reactor.fireSystemEvent('shutdown')
                    from mswin32 import tools
                    tools.kill_process('elisa_fork')
                    tools.kill_process('elisa')
                    import win32process
                    win32process.ExitProcess(0)
                else:
                    reactor.stop()

        if self.running:
            # stop the interface controller and the
            # managers
            self.info("Stopping interface controller")
            dfr = self.interface_controller.stop()
            dfr.addCallback(interface_controller_stopped)
            return dfr

        return defer.succeed(None)

    def _close_splash_screen(self):
        if self._splash != None:
            self._splash.destroy()
            
    def _on_copydata(self, hwnd, msg, wparam, lparam):
        import ctypes
        from ctypes import Structure, c_char, sizeof, addressof, memmove
        from ctypes.wintypes import ULONG, DWORD
        from mswin32.structures import COPYDATASTRUCT
        
        copydata = COPYDATASTRUCT()
        memmove(addressof(copydata), lparam, sizeof(copydata))
        
        class COMMANDLINEDATA(Structure):
            _fields_ = [
                        ("data", c_char * copydata.cbData)
                       ]
        
        commandlinedata = COMMANDLINEDATA()
        memmove(addressof(commandlinedata), copydata.lpData, sizeof(commandlinedata))
        frontend = self.interface_controller.frontends.values()[0]
        main = frontend.retrieve_controllers('/poblesec')[0]
        files = commandlinedata.data.split(" ")
        if len(files) > 0:
            main.play_file(files[0])
        
        return True
